#include "weedlog.h"

struct icmppacket {
  struct wl_iphdr {
#if __BYTE_ORDER == __LITTLE_ENDIAN
    u_char	ihl:4,
		version:4;
#elif __BYTE_ORDER == __BIG_ENDIAN
    u_char	version:4,
		ihl:4;
#else
#error	"Please fix <asm/byteorder.h>"
#endif
    u_char	tos;
    u_short	tot_len;
    u_short	id;
    u_short	frag_off;
    u_char	ttl;
    u_char	protocol;
    u_short	check;
    u_long	saddr;
    u_long	daddr;
  }ip;
  struct wl_icmphdr {
    u_char	type;
    u_char	code;
    u_short	checksum;
    union {
      struct {
	u_short	id;
	u_short	sequence;
      }echo;
      u_long	gateway;
      struct {
	u_short	__unused;
	u_short	mtu;
      }frag;
    }un;
  }icmp;
};


thread_function_t parse_icmp(void *argstruct) {
  int pktsize=0, gotpkt=0;
  char newbuf[512], suspip[24], psizebuf[24], tempflags[128], ipflags[128],
      saddress[82], daddress[88], pkttype[64], csums[64], miscstuff[200];
  struct parse_args *args = (struct parse_args *)argstruct;
  struct icmppacket *pkt = (struct icmppacket*)args->buffer;
  int verbose = args->verblev;
  time_t t;
  char *timestring = (char*)NULL;

  if(!pkt)
    goto picmp_end;

  if((timestring = (char*)malloc(32)) < 0) {
    fprintf(stderr, "%s[%i]: malloc: %s\n", progname, wlpid, strerror(errno));
    goto picmp_end;
  }
  if(dots) {
    if(dots > 1) {
      t = time(NULL);
      snprintf(timestring, 32, " (TS: %lu)", t);
    } else {
      struct tm *lt;
      t = time(NULL);
      lt = localtime(&t);
      if((strftime(timestring, 32, " (TS: %b %d %Y %T)", lt)) < 0) {
	fprintf(stderr, "%s[%i]: strftime: %s", progname, wlpid, strerror(errno));
      }
    }
  }

  if(pkt->ip.ihl != 5)
    snprintf(suspip, sizeof(suspip), " Suspicious IP Options: ");
  else
    snprintf(suspip, sizeof(suspip), " ");

  if((pktsize = ntohs(pkt->ip.tot_len)) > 1024)
    snprintf(psizebuf, sizeof(psizebuf), "%4.02f KBytes", (float)pktsize / 1024.0);
  else
    snprintf(psizebuf, sizeof(psizebuf), "%4.02f Bytes", (float)pktsize);

  if(pkt->ip.saddr) {
    struct in_addr s;
    s.s_addr = pkt->ip.saddr;
    if(skipdns) {
      snprintf(saddress, sizeof(saddress), "%s", inet_ntoa(s));
    } else {
      snprintf(saddress, sizeof(saddress), "%s[%s]",
	  hostlookup(pkt->ip.saddr), inet_ntoa(s));
    }
  } else {
    snprintf(saddress, sizeof(saddress), "no source found");
  }
  if(pkt->ip.daddr) {
    struct in_addr d;
    d.s_addr = pkt->ip.daddr;
    if(skipdns) {
      snprintf(daddress, sizeof(daddress), " to %s,", inet_ntoa(d));
    } else {
      snprintf(daddress, sizeof(daddress), " to %s[%s],",
	  hostlookup(pkt->ip.daddr), inet_ntoa(d));
    }
  } else {
    snprintf(daddress, sizeof(daddress), ",");
  }

  if(verbose >= 3) {
    snprintf(ipflags, sizeof(ipflags), "  IP_FLAGS:");
    if(pkt->ip.version > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s VER=%i", ipflags,
	  pkt->ip.version);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.protocol > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s PROTO=%i", ipflags,
	  pkt->ip.protocol);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.frag_off > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s FRAG_OFF=%i", ipflags,
	  pkt->ip.frag_off);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.tos > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s TOS=%i", ipflags,
	  pkt->ip.tos);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.id > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s ID=%i", ipflags,
	  pkt->ip.id);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.ttl > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s TTL=%i", ipflags,
	  pkt->ip.ttl);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
    if(pkt->ip.check > 0) {
      snprintf(tempflags, sizeof(tempflags), "%s CHECK=%i", ipflags,
	  pkt->ip.check);
      snprintf(ipflags, sizeof(ipflags), "%s", tempflags);
    }
  } else snprintf(ipflags, sizeof(ipflags), " ");

  if(pkt->icmp.type == 3) {
    if(pkt->icmp.code == 0) {
      snprintf(pkttype, sizeof(pkttype), "Network Unreachable (T:3 C:0)");
    } else if(pkt->icmp.code == 1) {
      snprintf(pkttype, sizeof(pkttype), "Host Unreachable (T:3 C:1)");
    } else if(pkt->icmp.code == 2) {
      snprintf(pkttype, sizeof(pkttype), "Protocol Unreachable (T:3 C:2)");
    } else if(pkt->icmp.code == 3) {
      snprintf(pkttype, sizeof(pkttype), "Port Unreachable (T:3 C:3)");
    } else if(pkt->icmp.code == 4) {
      snprintf(pkttype, sizeof(pkttype), "Fragmentation Needed (T:3 C:4)");
    } else if(pkt->icmp.code == 5) {
      snprintf(pkttype, sizeof(pkttype), "Source Route failed (T:3 C:5)");
    } else if(pkt->icmp.code == 6) {
      snprintf(pkttype, sizeof(pkttype), "Network Unknown (T:3 C:6)");
    } else if(pkt->icmp.code == 7) {
      snprintf(pkttype, sizeof(pkttype), "Host Unknown (T:3 C:7)");
    } else if(pkt->icmp.code == 8) {
      snprintf(pkttype, sizeof(pkttype), "Host Isolated (T:3 C:8)");
    } else if(pkt->icmp.code == 9) {
      snprintf(pkttype, sizeof(pkttype), "ICMP_NET_ANO (T:3 C:9)");
    } else if(pkt->icmp.code == 10) {
      snprintf(pkttype, sizeof(pkttype), "ICMP_HOST_ANO (T:3 C:10)");
    } else if(pkt->icmp.code == 11) {
      snprintf(pkttype, sizeof(pkttype), "Network Unreachable TOS (T:3 C:11)");
    } else if(pkt->icmp.code == 12) {
      snprintf(pkttype, sizeof(pkttype), "Host Unreachable TOS (T:3 C:12)");
    } else if(pkt->icmp.code == 13) {
      snprintf(pkttype, sizeof(pkttype), "Packet Filtered (T:3 C:13)");
    } else if(pkt->icmp.code == 14) {
      snprintf(pkttype, sizeof(pkttype), "Precedence violation (T:3 C:14)");
    } else if(pkt->icmp.code == 15) {
      snprintf(pkttype, sizeof(pkttype), "Precedence cut off (T:3 C:15)");
    } else {
      snprintf(pkttype, sizeof(pkttype), "Destination Unreachable (T:3)");
    }
    gotpkt = 1;
  }
  if(pkt->icmp.type == 4) {
    snprintf(pkttype, sizeof(pkttype), "Source Quench (T:4)");
    gotpkt = 1;
  }
  if(pkt->icmp.type == 5) {
    if(pkt->icmp.code == 0) {
      snprintf(pkttype, sizeof(pkttype), "Redirect Net (T:5 C:0)");
    } else if(pkt->icmp.code == 1) {
      snprintf(pkttype, sizeof(pkttype), "Redirect Host (T:5 C:1)");
    } else if(pkt->icmp.code == 2) {
      snprintf(pkttype, sizeof(pkttype), "Redirect Net for TOS (T:5 C:2)");
    } else if(pkt->icmp.code == 3) {
      snprintf(pkttype, sizeof(pkttype), "Redirect Host for TOS (T:5 C:3)");
    } else {
      snprintf(pkttype, sizeof(pkttype), "Redirect (T:5)");
    }
    gotpkt = 1;
  }
  if(pkt->icmp.type == 8) {
    snprintf(pkttype, sizeof(pkttype), "Echo/Ping (T:8)");
    gotpkt = 1;
  }
  if(pkt->icmp.type == 15) {
    snprintf(pkttype, sizeof(pkttype), "Info Request (T:15)");
    gotpkt = 1;
  }
  if(pkt->icmp.type == 11) {
    if(pkt->icmp.code == 0) {
      snprintf(pkttype, sizeof(pkttype), "Time Exceeded (TTL count exceeded) (T:11 C:0)");
    } else if(pkt->icmp.code == 1) {
      snprintf(pkttype, sizeof(pkttype),
	  "Time Exceeded (Fragment Reasses time exceeded) (T:11 C:1)");
    } else {
      snprintf(pkttype, sizeof(pkttype), "Time Exceeded (T:11)");
    }
    gotpkt = 1;
  }
  if(pkt->icmp.type == 12) {
    snprintf(pkttype, sizeof(pkttype), "Parameter Problem (T:12)");
    gotpkt = 1;
  }
  if(pkt->icmp.type == 13) {
    snprintf(pkttype, sizeof(pkttype), "Time Stamp (T:13)");
    gotpkt = 1;
  }
  if(pkt->icmp.type == 17) {
    snprintf(pkttype, sizeof(pkttype), "Address Mask Request (T:17)");
    gotpkt = 1;
  }
  if(verbose >= 1) {
    if(pkt->icmp.type == 0) {
      snprintf(pkttype, sizeof(pkttype), "Echo Reply (T:0)");
      gotpkt = 1;
    }
    if(pkt->icmp.type == 14) {
      snprintf(pkttype, sizeof(pkttype), "Time Stamp Reply (T:14)");
      gotpkt = 1;
    }
    if(pkt->icmp.type == 16) {
      snprintf(pkttype, sizeof(pkttype), "Info Reply (T:16)");
      gotpkt = 1;
    }
    if(pkt->icmp.type == 18) {
      snprintf(pkttype, sizeof(pkttype), "Address Mask Reply (T:18)");
      gotpkt = 1;
    }
  }
  if(!(gotpkt)) {
    if(verbose >=1) {
      snprintf(pkttype, sizeof(pkttype), "Type %u (code %u)", pkt->icmp.type,
	  pkt->icmp.code);
      gotpkt = 1;
    } else if(!((pkt->icmp.type == 0)
	|| (pkt->icmp.type == 14) || (pkt->icmp.type == 16)
	|| (pkt->icmp.type == 18))) {
      snprintf(pkttype, sizeof(pkttype), "Type %u (code %u)", pkt->icmp.type,
	  pkt->icmp.code);
      gotpkt = 1;
    }
  }

  if(verbose >= 2)
    snprintf(csums, sizeof(csums), "  Checksum: %u", pkt->icmp.checksum);
  else snprintf(csums, sizeof(csums), " ");

  if(verbose >= 2) {
    snprintf(miscstuff, sizeof(miscstuff), " ");
    if(pkt->icmp.un.echo.id) {
      snprintf(miscstuff, sizeof(miscstuff), "%s ECHO.ID: %u",
	  miscstuff, pkt->icmp.un.echo.id);
    }
    if(pkt->icmp.un.echo.sequence) {
      snprintf(miscstuff, sizeof(miscstuff), "%s ECHO.SEQ: %u",
	  miscstuff, pkt->icmp.un.echo.sequence);
    }
    if(pkt->icmp.un.frag.mtu) {
      snprintf(miscstuff, sizeof(miscstuff), "%s FRAG.MTU: %u",
	  miscstuff, pkt->icmp.un.frag.mtu);
    }
  } else snprintf(miscstuff, sizeof(miscstuff), " ");

  if(gotpkt) {
    if(dots) {
      snprintf(newbuf, sizeof(newbuf), "ICMP:%s%s%s packet from %s%s Size: %s"
	  "%s%s%s", timestring, suspip, pkttype, saddress, daddress, psizebuf,
	  csums, miscstuff, ipflags);
    } else {
      snprintf(newbuf, sizeof(newbuf), "ICMP:%s%s packet from %s%s Size: %s"
	  "%s%s%s", suspip, pkttype, saddress, daddress, psizebuf, csums,
	  miscstuff, ipflags);
    }
    print_log("%s\n", newbuf);
  }

picmp_end:
  if(timestring) free(timestring);
  if(args->buffer) free(args->buffer);
  if(argstruct) free(argstruct);
#ifdef USETHREADS
  if(getpid() != wlpid) {
    if(debug >= 4) print_log("debug: Child thread == %u, exiting thread\n",
	getpid());
    deactivate_thread();
    pthread_exit(NULL);
  } else {
    if(debug >= 4)
      print_log("debug: Child (%i) and Parent (%i) have the same pid?!\n",
	  getpid(), wlpid);
  }
#endif

  return 0;
}
